/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "zconf.tab.h"
#include "zconf.h"

static void conf(struct menu *menu);

static int indent = 1;
static int askold = 1;
static int conf_cnt;
static char line[128];

#if 0
static void printc(int ch)
{
	static int sep = 0;

	if (!sep) {
		putchar('[');
		sep = 1;
	} else if (ch)
		putchar('/');
	if (!ch) {
		putchar(']');
		putchar(' ');
		sep = 0;
	} else
		putchar(ch);
}
#endif

static void printo(char *o)
{
	static int sep = 0;

	if (!sep) {
		putchar('(');
		sep = 1;
	} else if (o) {
		putchar(',');
		putchar(' ');
	}
	if (!o) {
		putchar(')');
		putchar(' ');
		sep = 0;
	} else
		printf("%s", o);
}

static void strip(char *str)
{
	char *p = str;
	int l;

	while ((isspace(*p)))
		p++;
	l = strlen(p);
	if (p != str)
		memmove(str, p, l + 1);
	if (!l)
		return;
	p = str + l - 1;
	while ((isspace(*p)))
		*p-- = 0;
}

int conf_string(struct menu *menu)
{
	struct symbol *sym = menu->sym;
	char line[128];
	const char *def;

	while (1) {
		printf("%*s%s ", indent - 1, "", menu->prompt->prompt);
		printf("(%s) ", sym->name);
		def = sym_get_string_value(sym);
		if (sym_get_string_value(sym))
			printf("[%s] ", def);
		if (!sym_has_value(sym)) {
			printf("(NEW) ");
		} else {
			if (!askold) {
				printf("%s\n", def);
				return 0;
			}
		}
		fgets(line, 128, stdin);
		switch (line[0]) {
		case '\n':
			break;
		case '?':
			/* print help */
			if (line[1] == 0) {
				if (menu->sym->help)
					printf("\n%s", menu->sym->help);
				else
					printf("\nSorry, no help available for this option yet.\n\n");
				def = NULL;
				break;
			}
		default:
			line[strlen(line)-1] = 0;
			def = line;
		}
		if (def && sym_set_string_value(sym, def))
			return 0;
	}
}

static int conf_sym(struct menu *menu)
{
	struct symbol *sym = menu->sym;
	int type;
	tristate oldval, newval;

	while (1) {
		printf("%*s%s ", indent - 1, "", menu->prompt->prompt);
		if (sym->name)
			printf("(%s) ", sym->name);
		type = sym_get_type(sym);
		putchar('[');
		oldval = sym_get_tristate_value(sym);
		switch (oldval) {
		case no:
			putchar('N');
			break;
		case mod:
			putchar('M');
			break;
		case yes:
			putchar('Y');
			break;
		}
		if (oldval != no && sym_tristate_within_range(sym, no))
			printf("/n");
		if (oldval != mod && sym_tristate_within_range(sym, mod))
			printf("/m");
		if (oldval != yes && sym_tristate_within_range(sym, yes))
			printf("/y");
		if (sym->help)
			printf("/?");
		printf("] ");
		if (!sym_has_value(sym))
			printf("(NEW) ");
		else if (!askold) {
			printf("%s\n", sym_get_string_value(sym));
			return 0;
		}
		fgets(line, 128, stdin);
		strip(line);

		switch (line[0]) {
		case 'n':
		case 'N':
			newval = no;
			if (!line[1] || !strcmp(&line[1], "o"))
				break;
			continue;
		case 'm':
		case 'M':
			newval = mod;
			if (!line[1])
				break;
			continue;
		case 'y':
		case 'Y':
			newval = yes;
			if (!line[1] || !strcmp(&line[1], "es"))
				break;
			continue;
		case 0:
			newval = oldval;
			break;
		default:
			continue;
		}
		if (sym_set_tristate_value(sym, newval))
			return 0;
		if (sym->help)
			printf("\n%s", sym->help);
	}
}

static int conf_choice(struct menu *menu)
{
	struct symbol *sym, *def_sym;
	struct menu *cmenu, *def_menu;
	int type, len;
	bool is_new;
	char line[128];

	sym = menu->sym;
	type = sym_get_type(sym);
	is_new = !sym_has_value(sym);
	if (sym_is_changable(sym)) {
		conf_sym(menu);
		clear_sym_deps();
		sym_calc_value(sym);
		switch (sym_get_tristate_value(sym)) {
		case no:
			return 1;
		case mod:
			return 0;
		case yes:
			break;
		}
	} else {
		sym->def = sym->curr;
		if (S_TRI(sym->curr) == mod) {
			printf("%*s%s\n", indent - 1, "", menu_get_prompt(menu));
			return 0;
		}
	}

	while (1) {
		printf("%*s%s ", indent - 1, "", menu_get_prompt(menu));
		def_sym = sym_get_choice_value(sym);
		def_menu = NULL;
		for (cmenu = menu->list; cmenu; cmenu = cmenu->next) {
			if (!menu_is_visible(cmenu))
				continue;
			printo(menu_get_prompt(cmenu));
			if (cmenu->sym == def_sym)
				def_menu = cmenu;
		}
		printo(NULL);
		if (def_menu)
			printf("[%s] ", menu_get_prompt(def_menu));
		else {
			printf("\n");
			return 1;
		}
		if (is_new)
			printf("(NEW) ");
		else if (!askold) {
			printf("%s\n", menu_get_prompt(def_menu));
			return 1;
		}
		fgets(line, 128, stdin);
		if (line[0] == '?' && !line[1]) {
			if (menu->sym->help)
				printf("\n%s", menu->sym->help);
			continue;
		}
		if (line[0] != '\n') {
			len = strlen(line) - 1;
			line[len] = 0;

			def_menu = NULL;
			for (cmenu = menu->list; cmenu; cmenu = cmenu->next) {
				if (!cmenu->sym || !menu_is_visible(cmenu))
					continue;
				if (!strncmp(line, menu_get_prompt(cmenu), len)) {
					def_menu = cmenu;
					break;
				}
			}
		}
		if (def_menu) {
			sym_set_choice_value(sym, def_menu->sym);
			if (def_menu->list) {
				indent += 2;
				conf(def_menu->list);
				indent -= 2;
			}
			return 1;
		}
	}
}

static void conf(struct menu *menu)
{
	struct symbol *sym;
	struct menu *child;

	clear_sym_deps();
	if (!menu_is_visible(menu))
		return;

	sym = menu->sym;
	if (!sym) {
		if (menu->prompt)
			printf("%*c\n%*c %s\n%*c\n",
				indent, '*',
				indent, '*', menu->prompt->prompt,
				indent, '*');
		goto conf_childs;
	}

	if (sym_is_choice(sym)) {
		conf_choice(menu);
		if (S_TRI(sym->curr) != mod)
			return;
		goto conf_childs;
	}

	switch (sym->type) {
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		conf_string(menu);
		break;
	default:
		conf_sym(menu);
		break;
	}

conf_childs:
	if (sym)
		indent += 2;
	for (child = menu->list; child; child = child->next)
		conf(child);
	if (sym)
		indent -= 2;
}

static void check_conf(struct menu *menu)
{
	struct symbol *sym;
	struct menu *child;

	clear_sym_deps();
	if (!menu_is_visible(menu))
		return;

	sym = menu->sym;
	if (!sym)
		goto conf_childs;

	if (sym_is_choice(sym)) {
		if (!sym_has_value(sym)) {
			if (!conf_cnt++)
				printf("*\n* Restart config...\n*\n");
			conf(menu->parent);
		}
		if (sym_get_tristate_value(sym) != mod)
			return;
		goto conf_childs;
	}

	if (!sym_has_value(sym)) {
		if (!conf_cnt++)
			printf("*\n* Restart config...\n*\n");
		conf(menu->parent);
	}

conf_childs:
	for (child = menu->list; child; child = child->next)
		check_conf(child);
}

int main(int ac, char **av)
{
	int silent = 0;
	const char *name;

	if (ac > 1 && av[1][0] == '-') {
		switch (av[1][1]) {
		case 'o':
			askold = 0;
			break;
		case 's':
			silent = 1;
			break;
		case 'h':
		case '?':
			printf("%s [-o|-s] config\n", av[0]);
			exit(0);
		}
		name = av[2];
	} else
		name = av[1];
	conf_parse(name);
	conf_read(NULL);

	if (!silent)
		conf(&rootmenu);
	askold = 0;
	do {
		conf_cnt = 0;
		check_conf(&rootmenu);
	} while (conf_cnt);
	conf_write(NULL);
	return 0;
}
