%{
/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zconf.h"

#define printd(mask, fmt...) if (cdebug & (mask)) printf(fmt)

#define PRINTD		0x0001
#define DEBUG_PARSE	0x0002

int cdebug = PRINTD;

extern int zconflex(void);
static void zconferror(const char *err);

struct menu rootmenu;
static struct menu *current_menu;
static struct menu *last_menu;
static struct statement *first_stmt;
static struct expr *current_dep;

struct symbol *symbol_hash[257];

void reindent_menu(void);
struct statement *create_stmt(int token);
struct statement *add_stmt(int token, char *prompt, struct symbol *def, struct expr *dep);
struct menu *add_menu(struct symbol *sym);
struct expr *alloc_expr_and(struct expr *e1, struct expr *e2);

#define YYERROR_VERBOSE
%}

%union
{
	char *string;
	struct symbol *symbol;
	struct expr *expr;
	struct statement *stmt;
}

%token T_MAINMENU
%token T_MENU
%token T_ENDMENU
%token T_SOURCE
%token T_CHOICE
%token T_ENDCHOICE
%token T_COMMENT
%token T_CONFIG
%token T_HELP
%token <string> T_HELPTEXT
%token T_IF 304
%token T_DEPENDS
%token T_OPTIONAL
%token T_PROMPT 306
%token T_DEFAULT 307
%token T_TRISTATE, T_BOOLEAN, T_INT, T_HEX, T_STRING
%token <string> T_WORD 300, T_STRING
%token T_UNEQUAL 301

%left '|'
%left '&'
%left '=' T_UNEQUAL
%nonassoc '!'

%type <string> prompt
%type <symbol> symbol
%type <expr> expr, if_expr
%type <stmt> statement, menu_stmt, choice_stmt, source_stmt

%%
input:  block
;

block:	  /* empty */
	| block config_stmt
	| block choice_stmt
	| block menu_stmt
	| block source_stmt
	| block statement
;

config_stmt: T_CONFIG T_WORD '\n'
{
	struct symbol *sym = lookup_symbol($2, 0);
	printd(DEBUG_PARSE, "%s:%d:config %s\n", zconf_curname(), zconf_lineno(), $2);
	sym->flags |= SYMBOL_OPTIONAL;
	add_menu(sym);
}
	config_option_list
{
	printd(DEBUG_PARSE, "%s:%d:endconfig\n", zconf_curname(), zconf_lineno());
	if (current_menu->prompt && !current_menu->prompt->def) {
		switch (current_menu->sym->type) {
		case SYMBOL_INT:
		case SYMBOL_HEX:
		case SYMBOL_STRING:
			current_menu->prompt->def = alloc_symbol(NULL);
		}
	}
	reindent_menu();
};

config_option_list:
	  /* empty */
	| config_option_list config_option
{
};

config_option: T_DEPENDS expr '\n'
{
	current_dep = alloc_expr_and(current_dep, $2);
	printd(DEBUG_PARSE, "%s:%d:depends\n", zconf_curname(), zconf_lineno());
};

config_option: T_TRISTATE '\n'
{
	current_menu->sym->type = SYMBOL_TRISTATE;
	printd(DEBUG_PARSE, "%s:%d:tristate\n", zconf_curname(), zconf_lineno());
};

config_option: T_TRISTATE prompt if_expr '\n'
{
	current_menu->sym->type = SYMBOL_TRISTATE;
	current_menu->prompt = add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:tristate\n", zconf_curname(), zconf_lineno());
};

config_option: T_BOOLEAN '\n'
{
	current_menu->sym->type = SYMBOL_BOOLEAN;
	printd(DEBUG_PARSE, "%s:%d:boolean\n", zconf_curname(), zconf_lineno());
};

config_option: T_BOOLEAN prompt if_expr '\n'
{
	current_menu->sym->type = SYMBOL_BOOLEAN;
	current_menu->prompt = add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:boolean\n", zconf_curname(), zconf_lineno());
};

config_option: T_INT '\n'
{
	current_menu->sym->type = SYMBOL_INT;
	printd(DEBUG_PARSE, "%s:%d:int\n", zconf_curname(), zconf_lineno());
};

config_option: T_INT prompt if_expr '\n'
{
	current_menu->sym->type = SYMBOL_INT;
	current_menu->prompt = add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:int\n", zconf_curname(), zconf_lineno());
};

config_option: T_HEX '\n'
{
	current_menu->sym->type = SYMBOL_HEX;
	printd(DEBUG_PARSE, "%s:%d:hex\n", zconf_curname(), zconf_lineno());
};

config_option: T_HEX prompt if_expr '\n'
{
	current_menu->sym->type = SYMBOL_HEX;
	current_menu->prompt = add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:hex\n", zconf_curname(), zconf_lineno());
};

config_option: T_STRING '\n'
{
	current_menu->sym->type = SYMBOL_STRING;
	printd(DEBUG_PARSE, "%s:%d:string\n", zconf_curname(), zconf_lineno());
};

config_option: T_STRING prompt if_expr '\n'
{
	current_menu->sym->type = SYMBOL_STRING;
	current_menu->prompt = add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:string\n", zconf_curname(), zconf_lineno());
};

config_option: T_PROMPT prompt if_expr '\n'
{
	current_menu->prompt = add_stmt(T_PROMPT, $2, NULL, $3);
	printd(DEBUG_PARSE, "%s:%d:prompt\n", zconf_curname(), zconf_lineno());
};

config_option: T_DEFAULT symbol if_expr '\n'
{
	add_stmt(T_DEFAULT, NULL, $2, $3);
	printd(DEBUG_PARSE, "%s:%d:default\n", zconf_curname(), zconf_lineno());
};

config_option: T_HELP '\n'
{
	printd(DEBUG_PARSE, "%s:%d:help\n", zconf_curname(), zconf_lineno());
	zconf_starthelp();
}
	T_HELPTEXT
{
	current_menu->sym->help = $4;
};

choice_stmt: T_CHOICE '\n'
{
	struct symbol *sym;
	sym = alloc_symbol(NULL);
	sym->type = SYMBOL_BOOLEAN;
	sym->flags |= SYMBOL_CHOICE;
	sym->next = symbol_hash[256];
	symbol_hash[256] = sym;
	add_menu(sym);
	add_stmt('^', NULL, NULL, NULL);
	printd(DEBUG_PARSE, "%s:%d:choice\n", zconf_curname(), zconf_lineno());
}
	choice_option_list
{
	reindent_menu();
}
	choice_block T_ENDCHOICE '\n'
{
	struct menu *menu;
	struct symbol *def;

	while (!current_menu->sym || !(current_menu->sym->flags & SYMBOL_CHOICE))
		current_menu = current_menu->parent;

	//if (!(current_choice->flags & SYMBOL_OPTIONAL))
	//	if (!current_choice->def_value || current_choice->def_value == &symbol_no)
	//		current_choice->def_value = &symbol_yes;
	def = current_menu->prompt->def;
	if (def) for (menu = current_menu->list; menu; menu = menu->next) {
		if (!menu->sym || !menu->prompt)
			continue;
		//if (!def) {
		//	current_menu->stmt->def = menu->sym;
		//	break;
		//}
		if (!strncmp(def->name, menu->prompt->prompt, strlen(def->name))) {
			S_VAL(current_menu->sym->def) = menu->sym;
			break;
		}
	}

	current_menu = current_menu->parent;
	printd(DEBUG_PARSE, "%s:%d:endchoice\n", zconf_curname(), zconf_lineno());
};

choice_option_list:
	  /* empty */
	| choice_option_list choice_option
;

choice_option: T_DEPENDS expr '\n'
{
	current_dep = alloc_expr_and(current_dep, $2);
	printd(DEBUG_PARSE, "%s:%d:depends\n", zconf_curname(), zconf_lineno());
};

choice_option: T_PROMPT prompt '\n'
{
	current_menu->prompt = add_stmt(T_PROMPT, $2, NULL, NULL);
	printd(DEBUG_PARSE, "%s:%d:prompt\n", zconf_curname(), zconf_lineno());
};

choice_option: T_OPTIONAL '\n'
{
	current_menu->sym->flags |= SYMBOL_OPTIONAL;
	printd(DEBUG_PARSE, "%s:%d:optional\n", zconf_curname(), zconf_lineno());
};

choice_option: T_DEFAULT symbol '\n'
{
	add_stmt(T_DEFAULT, NULL, $2, NULL);
	//current_choice->stmt->def = $2;
	printd(DEBUG_PARSE, "%s:%d:default\n", zconf_curname(), zconf_lineno());
};

choice_option: T_HELP '\n'
{
	printd(DEBUG_PARSE, "%s:%d:help\n", zconf_curname(), zconf_lineno());
	zconf_starthelp();
}
	T_HELPTEXT
{
	current_menu->sym->help = $4;
};

choice_block:	  /* empty */
		| choice_block config_stmt
		| choice_block source_stmt
		| choice_block '\n'
;

menu_stmt: T_MENU prompt '\n'
{
	printd(DEBUG_PARSE, "%s:%d:menu\n", zconf_curname(), zconf_lineno());
	add_menu(NULL);
	current_menu->prompt = add_stmt(T_MENU, $2, NULL, NULL);
}
	stmt_dep_opt
{
	reindent_menu();
}
	menu_block T_ENDMENU '\n'
{
	while (!current_menu->prompt || current_menu->prompt->token != T_MENU)
		current_menu = current_menu->parent;
	current_menu = current_menu->parent;
	printd(DEBUG_PARSE, "%s:%d:endmenu\n", zconf_curname(), zconf_lineno());
};

menu_block:	  /* empty */
		| menu_block menu_stmt
		| menu_block choice_stmt
		| menu_block config_stmt
		| menu_block source_stmt
		| menu_block statement
;

statement: '\n'
{
	$$ = NULL;
};

statement: T_MAINMENU prompt '\n'
{
	printd(DEBUG_PARSE, "%s:%d:mainmenu\n", zconf_curname(), zconf_lineno());
	$$ = create_stmt(T_MAINMENU);
	$$->prompt = $2;
};

source_stmt: T_SOURCE prompt '\n'
{
	$$ = create_stmt(T_SOURCE);
	$$->prompt = $2;
	zconf_nextfile($2);
	printd(DEBUG_PARSE, "%s:%d:source %s\n", zconf_curname(), zconf_lineno(), $2);
};

statement: T_COMMENT prompt '\n'
{
	add_menu(NULL);
	current_menu->prompt = add_stmt(T_COMMENT, $2, NULL, NULL);
	printd(DEBUG_PARSE, "%s:%d:comment\n", zconf_curname(), zconf_lineno());
}
	stmt_dep_opt
{
	reindent_menu();
};

stmt_dep_opt:
	  /* empty */
	| T_DEPENDS expr '\n'
{
	current_dep = alloc_expr_and(current_dep, $2);
	printd(DEBUG_PARSE, "%s:%d:depends\n", zconf_curname(), zconf_lineno());
};

prompt:	  T_WORD
	| T_STRING
;

if_expr:  /* empty */			{ $$ = NULL; }
	| T_IF expr			{ $$ = $2; }
;

expr:	  symbol			{ $$ = alloc_expr_symbol($1); }
	| symbol '=' symbol		{ $$ = alloc_expr_comp('=', $1, $3); }
	| symbol T_UNEQUAL symbol	{ $$ = alloc_expr_comp(T_UNEQUAL, $1, $3); }
	| expr '|' expr			{ $$ = alloc_expr_two('|', $1, $3); }
	| expr '&' expr			{ $$ = alloc_expr_two('&', $1, $3); }
	| '(' expr ')'			{ $$ = $2; }
	| '!' expr			{ $$ = alloc_expr_one('!', $2); }
;

symbol:	  T_WORD	{ $$ = lookup_symbol($1, 0); free($1); }
	| T_STRING	{ $$ = lookup_symbol($1, 1); free($1); }
;

%%

void conf_parse(const char *name)
{
	zconf_initscan(name);

	current_menu = &rootmenu;
	rootmenu.prompt = create_stmt(T_PROMPT); // maybe T_IF?

	zconfparse();
	modules_sym = lookup_symbol("MODULES", 0);
}

static void zconferror(const char *err)
{
	printf("%s:%d:%s\n", zconf_curname(), zconf_lineno(), err);
	exit(1);
}

struct symbol *alloc_symbol(char *name)
{
	struct symbol *symbol;

	symbol = malloc(sizeof(*symbol));
	memset(symbol, 0, sizeof(*symbol));
	symbol->name = name;
	symbol->type = SYMBOL_UNKNOWN;
	symbol->flags = SYMBOL_NEW;

	return symbol;
}

struct symbol *lookup_symbol(char *name, int isconst)
{
	struct symbol *symbol;
	char *ptr;
	int hash = 0;

	//printf("lookup: %s -> ", name);
	if (name[0] && !name[1]) {
		switch (name[0]) {
		case 'y': return &symbol_yes;
		case 'm': return &symbol_mod;
		case 'n': return &symbol_no;
		}
	}
	for (ptr = name; *ptr; ptr++)
		hash += *ptr;
	hash &= 0xff;

	for (symbol = symbol_hash[hash]; symbol; symbol = symbol->next) {
		if (!strcmp(symbol->name, name)) {
			if ((isconst && symbol->flags & SYMBOL_CONST) ||
			    (!isconst && !(symbol->flags & SYMBOL_CONST))) {
				//printf("h:%p\n", symbol);
				return symbol;
			}
		}
	}

	symbol = alloc_symbol(strdup(name));
	if (isconst)
		symbol->flags |= SYMBOL_CONST;

	symbol->next = symbol_hash[hash];
	symbol_hash[hash] = symbol;

	//printf("n:%p\n", symbol);
	return symbol;
}

struct menu *add_menu(struct symbol *sym)
{
	struct menu *menu;

	menu = malloc(sizeof(*menu));
	memset(menu, 0, sizeof(*menu));
	menu->sym = sym;

	last_menu = current_menu;
	current_menu = menu;

	return menu;
}

void reindent_menu(void)
{
	struct menu *menu_block, *parent, **menup;
	struct statement *stmt;
	struct expr *dep;

	/* Find the enclosing menu block and add its dep to this menu entry */
	for (menu_block = last_menu; menu_block->parent; menu_block = menu_block->parent) {
		if ((menu_block->prompt && menu_block->prompt->token == T_MENU) ||
		    (menu_block->sym && sym_is_choice(menu_block->sym)))
			break;
	}

	if (menu_block->sym)
		dep = alloc_expr_symbol(menu_block->sym);
	else
		dep = copy_expr(E_EXPR(menu_block->prompt->visible));
	current_dep = alloc_expr_and(dep, current_dep);

	/* add dep to all stmt of this menu entry */
	for (stmt = first_stmt; stmt; stmt = stmt->next)
		E_EXPR(stmt->visible) = alloc_expr_and(copy_expr(current_dep), E_EXPR(stmt->visible));

	/* Find the parent menu entry */
	for (parent = last_menu; parent != menu_block; parent = parent->parent) {
		if (parent->sym && depends_on_symbol(current_dep, parent->sym))
			break;
	}

	if (current_menu->sym && parent->sym && sym_is_choice(parent->sym)) {
		current_menu->sym->flags |= SYMBOL_CHOICEVAL;
		add_stmt('^', NULL, parent->sym, NULL);
		if (current_menu->sym->type == SYMBOL_TRISTATE)
			parent->sym->type = SYMBOL_TRISTATE;

		stmt = sym_get_choice_stmt(parent->sym);
		stmt->dep = alloc_expr_one('^', stmt->dep);
		stmt->dep->right.sym = current_menu->sym;
	}

	/* only entries with correct prompt can have children (except root) */
	if (!parent->prompt || !parent->prompt->prompt)
		if (parent->parent)
			parent = parent->parent;

	/* append menu entry to its final parent */
	current_menu->parent = parent;
	for (menup = &parent->list; *menup; menup = &(*menup)->next)
		;
	*menup = current_menu;

	/* clean up */
	free_expr(current_dep);
	current_dep = NULL;
	first_stmt = NULL;
}

struct statement *create_stmt(int token)
{
	struct statement *stmt;

	stmt = malloc(sizeof(*stmt));
	memset(stmt, 0, sizeof(*stmt));
	stmt->token = token;
	stmt->file = current_file;
	stmt->lineno = zconf_lineno();
	if (!current_file->stmt)
		current_file->stmt = stmt;
	if (current_file->last_stmt)
		current_file->last_stmt->next_pos = stmt;
	current_file->last_stmt = stmt;

	return stmt;
}

struct statement *add_stmt(int token, char *prompt, struct symbol *def, struct expr *dep)
{
	struct statement *stmt = create_stmt(token);
	struct statement **stmtp;

	stmt->sym = current_menu->sym;
	stmt->prompt = prompt;
	stmt->def = def;
	E_EXPR(stmt->visible) = dep;

	/* append statement to the stmt list of symbol */
	if (stmt->sym) {
		for (stmtp = &stmt->sym->stmt; *stmtp; stmtp = &(*stmtp)->next)
			;
		*stmtp = stmt;
	}

	if (!first_stmt)
		first_stmt = stmt;

	return stmt;
}

struct expr *alloc_expr_and(struct expr *e1, struct expr *e2)
{
	if (!e1)
		return e2;
	return e2 ? alloc_expr_two('&', e1, e2) : e1;
}

void print_quoted_string(FILE *out, const char *str)
{
	const char *p;
	int len;

	putc('"', out);
	while ((p = strchr(str, '"'))) {
		len = p - str;
		if (len)
			fprintf(out, "%.*s", len, str);
		fputs("\\\"", out);
		str = p + 1;
	}
	fputs(str, out);
	putc('"', out);
}

void do_print_symbol(FILE *out, struct symbol *sym)
{
	struct statement *stmt;

	sym->flags |= SYMBOL_PRINTED;

	fprintf(out, "config %s\n", sym->name);
	switch (sym->type) {
	case SYMBOL_BOOLEAN:
		fputs("  boolean\n", out);
		break;
	case SYMBOL_TRISTATE:
		fputs("  tristate\n", out);
		break;
	case SYMBOL_STRING:
		fputs("  string\n", out);
		break;
	case SYMBOL_INT:
		fputs("  integer\n", out);
		break;
	case SYMBOL_HEX:
		fputs("  hex\n", out);
		break;
	}
#if 0
	if (!expr_is_yes(sym->dep)) {
		fputs("  depends ", out);
		fprint_expr(out, sym->dep, 0);
		fputc('\n', out);
	}
#endif
	for (stmt = sym->stmt; stmt; stmt = stmt->next) {
		switch (stmt->token) {
		case T_PROMPT:
			fputs("  prompt ", out);
			print_quoted_string(out, stmt->prompt);
			if (stmt->def) {
				fputc(' ', out);
				if (stmt->def->flags & SYMBOL_CONST)
					print_quoted_string(out, stmt->def->name);
				else
					fputs(stmt->def->name, out);
			}
			if (!expr_is_yes(E_EXPR(stmt->visible))) {
				fputs(" if ", out);
				fprint_expr(out, E_EXPR(stmt->visible), 0);
			}
			fputc('\n', out);
			break;
		case T_DEFAULT:
			fputs( "  default ", out);
			print_quoted_string(out, stmt->def->name);
			if (!expr_is_yes(E_EXPR(stmt->visible))) {
				fputs(" if ", out);
				fprint_expr(out, E_EXPR(stmt->visible), 0);
			}
			fputc('\n', out);
			break;
		case '^':
			fputs("  #choice value\n", out);
			break;
		default:
			fprintf(out, "  unknown stmt %d!\n", stmt->token);
			break;
		}
	}
	if (sym->help) {
		int len = strlen(sym->help);
		while (sym->help[--len] == '\n')
			sym->help[len] = 0;
		fprintf(out, "  help\n%s\n", sym->help);
	}
	fputc('\n', out);
}

void print_choice(FILE *out, struct symbol *sym)
{
	struct statement *stmt;
	struct expr *lst;

	for (stmt = sym->stmt; stmt; stmt = stmt->next)
		if (stmt->token == '^')
			break;
	if (!stmt)
		//warn?
		return;

	fputs("choice\n", out);
	if (stmt->prompt) {
		char *p = stmt->prompt;
		while (*p == ' ')
			p++;
		fputs("  prompt ", out);
		print_quoted_string(out, p);
		fputc('\n', out);
	}

	if (sym->flags & SYMBOL_CHOICE)
		fputs("  optional\n", out);

	if (!expr_is_yes(E_EXPR(stmt->visible))) {
		fputs("  depends ", out);
		fprint_expr(out, E_EXPR(stmt->visible), 0);
		fputc('\n', out);
	}
	if (stmt->def) {
		fputs("  default ", out);
		print_quoted_string(out, stmt->def->name);
		fputc('\n', out);
	}

	fputc('\n', out);

	for (lst = stmt->dep; lst; lst = lst->left.expr) {
		sym = lst->right.sym;
		if (sym != &symbol_no)
			do_print_symbol(out, sym);
	}

	fputs("endchoice\n\n", out);
}

void print_symbols(FILE *out, struct symbol *sym)
{
	if (sym->flags & SYMBOL_PRINTED)
		return;
	sym->flags |= SYMBOL_PRINTED;

	//if (sym->flags & SYMBOL_CHOICE) {
	//	print_choice(out, sym);
	//	return;
	//}
	do_print_symbol(out, sym);
}

void zconfdump(FILE *out)
{
	//struct file *file;
	struct statement *stmt;
	struct symbol *sym;
	struct menu *menu;

	print_type = 0;

	menu = rootmenu.list;
	while (menu) {
		if ((sym = menu->sym))
			print_symbols(out, sym);
		else {
			stmt = menu->prompt;
			switch (stmt->token) {
			case T_MAINMENU:
				fputs("\nmainmenu ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n", out);
				break;
			case T_COMMENT:
				fputs("\ncomment ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n", out);
				break;
			case T_MENU:
				fputs("\nmenu ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n", out);
				break;
			case T_ENDMENU:
				fputs("\nendmenu\n", out);
				break;
			case T_SOURCE:
				fputs("\nsource ", out);
				print_quoted_string(out, stmt->prompt);
				fputs("\n", out);
				break;
			}
			if (!expr_is_yes(E_EXPR(stmt->visible))) {
				fputs("  depends ", out);
				fprint_expr(out, E_EXPR(stmt->visible), 0);
				fputc('\n', out);
			}
			fputs("\n", out);
		}

		if (menu->list)
			menu = menu->list;
		else if (menu->next)
			menu = menu->next;
		else while ((menu = menu->parent)) {
			if (menu->next) {
				menu = menu->next;
				break;
			}
		}
	}
}
